# pylint: disable=no-name-in-module

import csv
import os
import numpy as np

from PIL import Image
from tqdm import trange
from d3rlpy.dataset import MDPDataset


def convert_ndarray_to_image(ndarray):
    # convert channel-fist to channel-last
    if ndarray.shape[0] == 1:
        array = ndarray[0]
    else:
        array = np.transpose(ndarray, [1, 2, 0])
    return Image.fromarray(array)


def convert_image_to_ndarray(image):
    array = np.asarray(image)
    # fix channel-first shape
    if image.mode == 'L':
        array = array.reshape((1, *array.shape))
    else:
        array = array.transpose([2, 0, 1])
    return array


def export_mdp_dataset_as_csv(dataset, fname, relative_path=False):
    if len(dataset.get_observation_shape()) > 1:
        # image observation
        export_image_observation_dataset_as_csv(dataset, fname, relative_path)
    else:
        # vector observation
        export_vector_observation_dataset_as_csv(dataset, fname)


def _save_image_files(dataset, dir_path):
    data_size = dataset.observations.shape[0]
    for i in trange(data_size, desc='saving images'):
        image = convert_ndarray_to_image(dataset.observations[i])
        image_path = os.path.join(dir_path, 'observation_%d.png' % i)
        image.save(image_path, quality=100)


def export_image_observation_dataset_as_csv(dataset, fname, relative_path):
    action_size = dataset.get_action_size()

    # prepare image directory
    csv_file_name = os.path.basename(fname)
    image_dir_name = csv_file_name.split('.')[0] + '_images'
    image_dir_path = os.path.join(os.path.dirname(fname), image_dir_name)
    os.makedirs(image_dir_path, exist_ok=True)

    # save image files
    _save_image_files(dataset, image_dir_path)

    file_name = 'observation_%d.png'
    if relative_path:
        image_path = os.path.join(image_dir_name, file_name)
    else:
        image_path = file_name

    with open(fname, 'w') as file:
        writer = csv.writer(file)

        # write header
        header = ['episode', 'observation:0']
        if dataset.is_action_discrete():
            header += ['action:0']
        else:
            header += ['action:%d' % i for i in range(action_size)]
        header += ['reward']
        writer.writerow(header)

        count = 0
        for i, episode in enumerate(dataset.episodes):
            # prepare data to write
            for j in range(episode.observations.shape[0]):
                row = []
                row.append(i)

                # add image path
                row.append(image_path % count)
                count += 1

                row += episode.actions[j].reshape(-1).tolist()
                row.append(episode.rewards[j])
                writer.writerow(row)


def export_vector_observation_dataset_as_csv(dataset, fname):
    observation_size = dataset.get_observation_shape()[0]
    action_size = dataset.get_action_size()

    with open(fname, 'w') as file:
        writer = csv.writer(file)

        # write header
        header = ['episode']
        header += ['observation:%d' % i for i in range(observation_size)]
        if dataset.is_action_discrete():
            header += ['action:0']
        else:
            header += ['action:%d' % i for i in range(action_size)]
        header += ['reward']
        writer.writerow(header)

        for i, episode in enumerate(dataset.episodes):
            # prepare data to write
            observations = np.asarray(episode.observations)
            episode_length = observations.shape[0]
            actions = np.asarray(episode.actions).reshape(episode_length, -1)
            rewards = episode.rewards.reshape(episode_length, 1)
            episode_ids = np.full([episode_length, 1], i)

            # write episode
            rows = np.hstack([episode_ids, observations, actions, rewards])
            writer.writerows(rows)


def import_csv_as_mdp_dataset(fname, image=False):
    if image:
        return import_csv_as_image_observation_dataset(fname)
    return import_csv_as_vector_observation_dataset(fname)


def _load_image(path):
    image = Image.open(path)

    # resize image to (84, 84)
    if image.size != (84, 84):
        image = image.resize((84, 84), Image.BICUBIC)

    return image


def import_csv_as_image_observation_dataset(fname):
    with open(fname, 'r') as file:
        reader = csv.reader(file)
        rows = [row for row in reader]

        # check header
        header = rows[0]
        _validate_csv_header(header)

        # get action size
        action_size = _get_action_size_from_header(header)

        data_size = len(rows) - 1

        observations = []
        actions = []
        rewards = []
        terminals = []
        for i, row in enumerate(rows[1:]):
            episode_id = row[0]

            # load image
            image = _load_image(os.path.join(os.path.dirname(fname), row[1]))

            # convert PIL.Image to ndarray
            array = convert_image_to_ndarray(image)

            observations.append(array)

            # get action columns
            actions.append(list(map(float, row[2:2 + action_size])))

            # get reward column
            rewards.append(float(row[-1]))

            if i == data_size - 1 or episode_id != rows[i + 2][0]:
                terminals.append(1)
            else:
                terminals.append(0)

        # convert list to ndarray
        observations = np.array(observations, dtype=np.uint8)
        actions = np.array(actions)
        rewards = np.array(rewards, dtype=np.float32)
        terminals = np.array(terminals, dtype=np.float32)

        dataset = MDPDataset(observations=observations,
                             actions=actions,
                             rewards=rewards,
                             terminals=terminals)

    return dataset


def import_csv_as_vector_observation_dataset(fname):
    with open(fname, 'r') as file:
        reader = csv.reader(file)
        rows = [row for row in reader]

        # get observation shape
        header = rows[0]
        _validate_csv_header(header)

        # retrieve data section
        csv_data = np.array(rows[1:], dtype=np.float32)

        # get observation columns
        observation_size = _get_observation_size_from_header(header)
        observation_last_index = observation_size + 1
        observations = csv_data[:, 1:observation_last_index]

        # get action columns
        action_size = _get_action_size_from_header(header)
        action_last_index = observation_last_index + action_size
        actions = csv_data[:, observation_last_index:action_last_index]

        # get reward column
        rewards = csv_data[:, -1]

        # make terminal flags
        episode_ids = csv_data[:, 0]
        terminals = np.zeros_like(episode_ids)
        for i, episode_id in enumerate(episode_ids):
            if i + 1 == len(episode_ids) or episode_id != episode_ids[i + 1]:
                terminals[i] = 1.0

        dataset = MDPDataset(observations=observations,
                             actions=actions,
                             rewards=rewards,
                             terminals=terminals)

    return dataset


def _validate_csv_header(header):
    assert header[0] == 'episode', "column=0 must be 'episode'"

    # check observation section
    index = 1
    observation_index = 0
    while header[index].find('action') == -1:
        ref_name = 'observation:%d' % observation_index
        message = "column=%d must be '%s'" % (index, ref_name)
        assert header[index] == ref_name, message
        index += 1
        observation_index += 1

    # check action section
    action_index = 0
    while header[index] != 'reward':
        ref_name = 'action:%d' % action_index
        message = "column=%d must be '%s'" % (index, ref_name)
        assert header[index] == ref_name, message
        index += 1
        action_index += 1


def _get_observation_size_from_header(header):
    size = 0
    for column in header:
        if column.find('observation') > -1:
            size += 1
    return size


def _get_action_size_from_header(header):
    size = 0
    for column in header:
        if column.find('action') > -1:
            size += 1
    return size



# pylint: disable=no-member, no-self-use

from datetime import datetime
from werkzeug.exceptions import NotFound
from ..database import db


class BaseModel:
    created_at = db.Column(db.DateTime, nullable=True, default=datetime.now)
    updated_at = db.Column(db.DateTime,
                           nullable=True,
                           default=datetime.now,
                           onupdate=datetime.now)

    def delete(self):
        db.session.delete(self)
        db.session.commit()

    @classmethod
    def get(cls, model_id, raise_404=False):
        data = db.session.query(cls).get(model_id)
        if not data and raise_404:
            raise NotFound()
        return data

    @classmethod
    def create_query(cls):
        return db.session.query(cls)

    def update(self):
        db.session.commit()



# pylint: disable=no-member, too-many-instance-attributes

import os
import re
import json
import shutil
import numpy as np
from d3rlpy.algos import get_algo

from .base import BaseModel
from ..config import get_config
from ..database import db, ma
from ..async_helper import is_running, dispatch, kill
from ..tasks import train


def _get_metrics(log_path):
    metrics = {}
    pattern = re.compile('^.*.csv$')
    file_names = os.listdir(log_path)
    csv_names = filter(lambda name: re.match(pattern, name), file_names)
    for file_name in csv_names:
        path = os.path.join(log_path, file_name)
        name = file_name[:-4]
        data = np.loadtxt(path, delimiter=',')
        if len(data.shape) == 1:
            data = data.reshape((1, -1))
        data[np.isnan(data)] = 0.0
        metrics[name] = data[:, 2]
    return metrics


class Experiment(db.Model, BaseModel):
    __tablename__ = 'experiments'
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer, primary_key=True)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id'))
    name = db.Column(db.String(100))
    log_name = db.Column(db.String(100))
    config = db.Column(db.Text)
    is_active = db.Column(db.Boolean, default=True)

    def __init__(self, project_id, name, log_name, config):
        self.project_id = project_id
        self.name = name
        self.log_name = log_name
        self.config = config

    def __repr__(self):
        return '<Experiment {}:{}>'.format(self.id, self.name)

    @classmethod
    def create(cls, project_id, name, log_name, config):
        experiment = Experiment(project_id, name, log_name, config)
        db.session.add(experiment)
        db.session.commit()
        return experiment

    def delete(self):
        super().delete()
        # remove log directory
        if os.path.exists(self.get_log_path()):
            shutil.rmtree(self.get_log_path())

    def get_metrics(self):
        if not os.path.exists(self.get_log_path()):
            return {}
        return _get_metrics(self.get_log_path())

    def get_current_status(self):
        return is_running(self.id)

    def start_training(self):
        dispatch(self.id,
                 train,
                 algo_name=self.project.algorithm,
                 params=json.loads(self.config),
                 dataset_path=self.project.dataset.get_dataset_path(),
                 experiment_name=self.log_name,
                 logdir=get_config('LOG_DIR'))

    def cancel_training(self):
        kill(self.id)
        self.is_active = False
        self.update()

    def get_log_path(self):
        return os.path.join(get_config('LOG_DIR'), self.log_name)

    def save_policy(self, path, epoch, as_onnx):
        params_path = os.path.join(self.get_log_path(), 'params.json')
        model_path = os.path.join(self.get_log_path(), 'model_%d.pt' % epoch)

        if not os.path.exists(model_path):
            raise ValueError('%s does not exist.' % model_path)

        # initialize algorithm from json file
        algorithm_name = self.project.algorithm
        is_discrete = self.project.dataset.is_discrete
        algo = get_algo(algorithm_name, is_discrete).from_json(params_path)

        # load model parameters
        algo.load_model(model_path)

        # save TorchScript policy
        algo.save_policy(path, as_onnx)


class ExperimentSchema(ma.SQLAlchemySchema):
    class Meta:
        model = Experiment
        fields = ('id', 'project_id', 'name', 'log_name', 'config',
                  'is_active', 'updated_at')

    created_at = ma.DateTime('%Y-%m-%dT%H:%M:%S+09:00')
    updated_at = ma.DateTime('%Y-%m-%dT%H:%M:%S+09:00')



import os

if os.getenv('MINERVA_DIR'):
    ROOT_DIR = os.getenv('MINERVA_DIR')
else:
    ROOT_DIR = os.path.join(os.getenv('HOME'), '.minerva')

DATASET_DIR = os.path.join(ROOT_DIR, 'dataset')
DATABASE_PATH = os.path.join(ROOT_DIR, 'database.db')
MIGRATION_DIR = os.path.join(ROOT_DIR, 'migrations')
LOG_DIR = os.path.join(ROOT_DIR, 'train_logs')
TMP_DIR = '/tmp/minerva'
UPLOAD_DIR = '/tmp/minerva'

config = {
    'SQLALCHEMY_DATABASE_URI': 'sqlite:///%s' % DATABASE_PATH,
    'SQLALCHEMY_TRACK_MODIFICATIONS': False,
    'MAX_CONTENT_LENGTH': 10 * 1024 * 1024 * 1024,
    'SEND_FILE_MAX_AGE_DEFAULT': 0
}


def prepare_directory():
    if not os.path.exists(ROOT_DIR):
        os.makedirs(ROOT_DIR)

    if not os.path.exists(DATASET_DIR):
        os.makedirs(DATASET_DIR)

    if not os.path.exists(UPLOAD_DIR):
        os.makedirs(UPLOAD_DIR)

    if not os.path.exists(LOG_DIR):
        os.makedirs(LOG_DIR)


def get_config(key):
    paths = {
        'ROOT_DIR': ROOT_DIR,
        'DATASET_DIR': DATASET_DIR,
        'DATABASE_PATH': DATABASE_PATH,
        'LOG_DIR': LOG_DIR,
        'TMP_DIR': TMP_DIR,
        'UPLOAD_DIR': UPLOAD_DIR,
        'MIGRATION_DIR': MIGRATION_DIR
    }
    return paths[key]



from flask import request, jsonify
from sqlalchemy import desc


def update(cls, schema):
    def _func(model_id):
        model = cls.get(model_id, raise_404=True)
        json_data = request.get_json()
        model.name = json_data['name']
        model.update()
        return jsonify(schema().dump(model))

    return _func


def get_all(cls, schema):
    def _func():
        models = cls.create_query().order_by(desc(cls.id)).all()
        model_schema = schema(many=True)
        return jsonify({
            cls.__name__.lower() + 's': model_schema.dump(models),
            'total': len(models)
        })

    return _func


def get(cls, schema):
    def _func(model_id):
        model = cls.get(model_id, raise_404=True)
        return jsonify(schema().dump(model))

    return _func


def delete(cls):
    def _func(model_id):
        model = cls.get(model_id, raise_404=True)
        model.delete()
        return jsonify({})

    return _func


def generate_for_model(app, cls, schema):
    # get all endpoint
    app.add_url_rule('/', 'get_all', get_all(cls, schema), methods=['GET'])

    # get endpoint
    app.add_url_rule('/<model_id>', 'get', get(cls, schema), methods=['GET'])

    # update endpoint
    app.add_url_rule('/<model_id>',
                     'update',
                     update(cls, schema),
                     methods=['PUT'])

    # delete endpoint
    app.add_url_rule('/<model_id>', 'delete', delete(cls), methods=['DELETE'])



# pylint: disable=unidiomatic-typecheck

import json

from collections import defaultdict
from d3rlpy.gpu import get_gpu_count
from flask import Blueprint, jsonify
from ..models.experiment import Experiment, ExperimentSchema
from .project import _process_metrics

system_route = Blueprint('system', __name__)


def _get_device_id_and_data(experiment):
    # identify device
    config = json.loads(experiment.config)
    device_id = config['use_gpu'] if 'use_gpu' in config else None

    # make response data
    data = ExperimentSchema().dump(experiment)
    # update status
    _process_metrics(experiment, data)

    return device_id, data


@system_route.route('/status', methods=['GET'])
def get_system_status():
    n_gpus = get_gpu_count()

    # get all active experiments
    experiments = Experiment.create_query().filter(Experiment.is_active).all()

    gpu_jobs = defaultdict(list)
    cpu_jobs = []
    for experiment in experiments:
        device_id, data = _get_device_id_and_data(experiment)
        if type(device_id) == int:
            gpu_jobs[device_id].append(data)
        else:
            cpu_jobs.append(data)

    res = {
        'gpu': {
            'total': n_gpus,
            'jobs': gpu_jobs
        },
        'cpu': {
            'jobs': cpu_jobs
        }
    }

    return jsonify(res)



import time

from multiprocessing import Process, Queue

RUNNING_PROCESSES = {}


def _child_process(func, queue, args, kwargs):
    ret = func(*args, **kwargs)
    queue.put(ret)


def dispatch(uid, func, *args, **kwargs):
    queue = Queue()
    process = Process(target=_child_process,
                      args=(func, queue, args, kwargs),
                      daemon=True)
    process.start()
    RUNNING_PROCESSES[uid] = (process, queue)


def get(uid):
    _, queue = RUNNING_PROCESSES[uid]
    return queue.get()


def is_running(uid):
    if uid not in RUNNING_PROCESSES:
        return False
    process, _ = RUNNING_PROCESSES[uid]
    return process.is_alive()


def kill(uid):
    if uid not in RUNNING_PROCESSES:
        return
    process, _ = RUNNING_PROCESSES[uid]
    while process.is_alive():
        process.terminate()
        time.sleep(0.1)



# pylint: disable=unused-import

import os
import shutil
import click
import numpy as np
from flask import Flask, redirect
from flask.json import JSONEncoder

import minerva.models
import minerva.database as database
from minerva._version import __version__
from minerva.config import DATABASE_PATH, ROOT_DIR, MIGRATION_DIR
from minerva.config import config, prepare_directory
from minerva.database import init_db, db
from minerva.controllers import dataset_route
from minerva.controllers import project_route
from minerva.controllers import system_route

static_path = os.path.join(os.path.dirname(__file__), 'dist')
app = Flask(__name__, static_folder=static_path)
for name, val in config.items():
    app.config[name] = val
app.url_map.strict_slashes = False


# enable to return ndarray in json
class CustomJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        return JSONEncoder.default(self, o)


app.json_encoder = CustomJSONEncoder

# initialize database
init_db(app)

# create minerva directory
prepare_directory()

# API endpoints
app.register_blueprint(dataset_route, url_prefix='/api/datasets')
app.register_blueprint(project_route, url_prefix='/api/projects')
app.register_blueprint(system_route, url_prefix='/api/system')


# proxy
@app.route('/', defaults={'path': ''})
@app.route('/<path:path>')
def send_file(path):
    if path in ['', '/']:
        return redirect('/projects')
    if path.find('favicon.ico') > -1:
        path = 'favicon.ico'
    elif path.find('.js') == -1:
        path = 'index.html'
    return app.send_static_file(path)


@click.group()
def cli():
    print("MINERVA command line interface (Version %s)" % __version__)


@cli.command(short_help='Start MINERVA server.')
@click.option('--host', '-h', default='0.0.0.0', help='Host IP address.')
@click.option('--port', '-p', default=9000, help='Port number.')
@click.option("--debug", is_flag=True, help="Start as debug mode.")
def run(host, port, debug):
    # create databse if not exists
    if not os.path.exists(DATABASE_PATH):
        with app.app_context():
            db.create_all()
            database.init_migration(MIGRATION_DIR)
            database.create_migration(MIGRATION_DIR)

    # start server
    app.run(debug=debug, host=host, port=int(port))


@cli.command(short_help='Initialize database.')
def create_db():
    with app.app_context():
        db.create_all()


@cli.command(short_help='Upgrade database schema.')
def upgrade_db():
    with app.app_context():
        database.create_migration(MIGRATION_DIR)
        database.upgrade_db(MIGRATION_DIR)


@cli.command(short_help='Rollback database schema.')
def downgrade_db():
    with app.app_context():
        database.downgrade_db(MIGRATION_DIR)


@cli.command(short_help='Delete all data from the disk.')
def clean():
    shutil.rmtree(ROOT_DIR)


if __name__ == '__main__':
    cli()



# pylint: disable=no-member, too-many-instance-attributes, no-name-in-module

import os
from d3rlpy.dataset import MDPDataset

from ..config import get_config
from ..database import db, ma
from .base import BaseModel
from .project import Project


class Dataset(db.Model, BaseModel):
    __tablename__ = 'datasets'
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(100))
    file_name = db.Column(db.String(100))
    episode_size = db.Column(db.Integer)
    step_size = db.Column(db.Integer)
    data_size = db.Column(db.Integer)
    is_image = db.Column(db.Boolean)
    is_discrete = db.Column(db.Boolean)
    statistics = db.Column(db.Text)

    projects = db.relationship(Project, backref='dataset')

    def __init__(self, name, file_name, episode_size, step_size, data_size,
                 is_image, is_discrete, statistics):
        self.name = name
        self.file_name = file_name
        self.episode_size = episode_size
        self.step_size = step_size
        self.data_size = data_size
        self.is_image = is_image
        self.is_discrete = is_discrete
        self.statistics = statistics

    def __repr__(self):
        return '<Dataset {}:{}>'.format(self.id, self.name)

    @classmethod
    def create(cls, name, file_name, episode_size, step_size, data_size,
               is_image, is_discrete, statistics):
        dataset = Dataset(name, file_name, episode_size, step_size, data_size,
                          is_image, is_discrete, statistics)
        db.session.add(dataset)
        db.session.commit()
        return dataset

    def delete(self):
        # remove dataset file
        path = os.path.join(get_config('DATASET_DIR'), self.file_name)
        if os.path.exists(path):
            os.remove(path)
        super().delete()

    def load_mdp_dataset(self):
        path = self.get_dataset_path()
        mdp_dataset = MDPDataset.load(path)
        return mdp_dataset

    def get_dataset_path(self):
        return os.path.join(get_config('DATASET_DIR'), self.file_name)


class DatasetSchema(ma.SQLAlchemySchema):
    class Meta:
        model = Dataset
        fields = ('id', 'name', 'file_name', 'episode_size', 'step_size',
                  'data_size', 'is_image', 'is_discrete', 'statistics',
                  'created_at', 'updated_at')

    created_at = ma.DateTime('%Y-%m-%dT%H:%M:%S+09:00')
    updated_at = ma.DateTime('%Y-%m-%dT%H:%M:%S+09:00')



# pylint: disable=no-member, too-many-instance-attributes

from .base import BaseModel
from .experiment import Experiment
from ..database import db, ma


class Project(db.Model, BaseModel):
    __tablename__ = 'projects'
    __table_args__ = {'sqlite_autoincrement': True}
    id = db.Column(db.Integer, primary_key=True)
    dataset_id = db.Column(db.Integer, db.ForeignKey('datasets.id'))
    name = db.Column(db.String(100))
    algorithm = db.Column(db.String(100))

    experiments = db.relationship(Experiment, backref='project')

    def __init__(self, dataset_id, name, algorithm):
        self.dataset_id = dataset_id
        self.name = name
        self.algorithm = algorithm

    def __repr__(self):
        return '<Project {}:{}>'.format(self.id, self.name)

    @classmethod
    def create(cls, dataset_id, name, algorithm):
        project = Project(dataset_id, name, algorithm)
        db.session.add(project)
        db.session.commit()
        return project


class ProjectSchema(ma.SQLAlchemySchema):
    class Meta:
        model = Project

    id = ma.Integer(required=True)
    dataset_id = ma.Integer(required=True)
    name = ma.String(required=True)
    algorithm = ma.String(required=True)
    dataset = ma.Nested('minerva.models.dataset.DatasetSchema')
    created_at = ma.DateTime('%Y-%m-%dT%H:%M:%S+09:00')
    updated_at = ma.DateTime('%Y-%m-%dT%H:%M:%S+09:00')



import json
import uuid
import os
from flask import Blueprint, request, jsonify, send_file
from werkzeug.exceptions import NotFound
from sqlalchemy import desc

from .generator import generate_for_model
from ..config import get_config
from ..models.project import Project, ProjectSchema
from ..models.experiment import Experiment, ExperimentSchema

project_route = Blueprint('project', __name__)

generate_for_model(project_route, Project, ProjectSchema)


@project_route.route('/', methods=['POST'])
def create_project():
    json_data = request.get_json()
    dataset_id = json_data['dataset_id']
    name = json_data['name']
    algorithm = json_data['algorithm']
    project = Project.create(dataset_id, name, algorithm)
    return jsonify(ProjectSchema().dump(project))


def _process_metrics(experiment, data):
    current_status = experiment.get_current_status()
    if experiment.is_active and not current_status:
        experiment.is_active = False
        experiment.update()
    data['metrics'] = experiment.get_metrics()


@project_route.route('/<project_id>/experiments', methods=['GET'])
def get_all_experiments(project_id):
    experiments = Experiment.create_query()\
        .filter(Experiment.project_id == int(project_id))\
        .order_by(desc(Experiment.id))\
        .all()

    experiment_schema = ExperimentSchema(many=True)
    data = experiment_schema.dump(experiments)

    # update status
    for experiment, json_data in zip(experiments, data):
        _process_metrics(experiment, json_data)

    return jsonify({'experiments': data, 'total': len(experiments)})


@project_route.route('/<project_id>/experiments/<experiment_id>',
                     methods=['GET'])
def get_experiment(project_id, experiment_id):
    experiment = Experiment.get(experiment_id, raise_404=True)
    if experiment.project_id != int(project_id):
        return NotFound()

    data = ExperimentSchema().dump(experiment)

    _process_metrics(experiment, data)

    return jsonify(data)


@project_route.route('/<project_id>/experiments', methods=['POST'])
def create_experiment(project_id):
    json_data = request.get_json()
    name = json_data['name']
    config = json_data['config']
    log_name = str(uuid.uuid1())
    experiment = Experiment.create(project_id, name, log_name,
                                   json.dumps(config))

    # start training
    experiment.start_training()

    data = ExperimentSchema().dump(experiment)
    data['metrics'] = {}

    return jsonify(data)


@project_route.route('/<project_id>/experiments/<experiment_id>',
                     methods=['PUT'])
def update_experiment(project_id, experiment_id):
    experiment = Experiment.get(experiment_id, raise_404=True)
    if experiment.project_id != int(project_id):
        return NotFound()

    json_data = request.get_json()
    experiment.name = json_data['name']
    experiment.update()

    data = ExperimentSchema().dump(experiment)
    data['metrics'] = {}

    return jsonify(data)


@project_route.route('/<project_id>/experiments/<experiment_id>',
                     methods=['DELETE'])
def delete_experiment(project_id, experiment_id):
    experiment = Experiment.get(experiment_id, raise_404=True)
    if experiment.project_id != int(project_id):
        return NotFound()
    experiment.delete()
    return jsonify({})


@project_route.route('/<project_id>/experiments/<experiment_id>/cancel',
                     methods=['POST'])
def cancel_experiment(project_id, experiment_id):
    experiment = Experiment.get(experiment_id, raise_404=True)
    if experiment.project_id != int(project_id):
        return NotFound()

    # kill training process
    experiment.cancel_training()

    data = ExperimentSchema().dump(experiment)

    _process_metrics(experiment, data)

    return jsonify(data)


@project_route.route('/<project_id>/experiments/<experiment_id>/download',
                     methods=['GET'])
def download_policy(project_id, experiment_id):
    experiment = Experiment.get(experiment_id, raise_404=True)
    if experiment.project_id != int(project_id):
        return NotFound()

    epoch = int(request.args.get('epoch'))

    model_format = request.args.get('format')
    if model_format == 'torchscript':
        as_onnx = False
        file_name = 'policy.pt'
    elif model_format == 'onnx':
        as_onnx = True
        file_name = 'policy.onnx'
    else:
        raise ValueError('format should be torchscript or onnx.')

    # save greedy-policy as TorchScript or ONNX
    save_path = os.path.join(get_config('TMP_DIR'), file_name)
    experiment.save_policy(save_path, epoch, as_onnx)

    return send_file(save_path,
                     as_attachment=True,
                     attachment_filename=file_name,
                     mimetype='application/octet-stream')



# pylint: disable=unused-argument

import flask_migrate

from flask_sqlalchemy import SQLAlchemy
from flask_marshmallow import Marshmallow
from flask_migrate import Migrate


# protect sqlite internal tables from migrations
def skip_internal_tables(obj, name, type_, reflexted, compare_to):
    if type_ == 'table' and name == 'sqlite_sequence':
        return False
    return True


db = SQLAlchemy()
ma = Marshmallow()
migrate = Migrate(include_object=skip_internal_tables)


def init_db(app):
    db.init_app(app)
    ma.init_app(app)
    migrate.init_app(app, db)


def init_migration(directory):
    flask_migrate.init(directory)


def create_migration(directory):
    flask_migrate.migrate(directory)


def upgrade_db(directory):
    flask_migrate.upgrade(directory)


def downgrade_db(directory):
    flask_migrate.downgrade(directory)



import os

from setuptools import setup


# get __version__ variable
here = os.path.abspath(os.path.dirname(__file__))
exec(open(os.path.join(here, 'minerva', '_version.py')).read())

setup(name="minerva-ui",
      version=__version__,
      description="An out-of-the-box GUI tool for data-driven deep reinforcement learning",
      long_description=open("README.md").read(),
      long_description_content_type="text/markdown",
      url="https://github.com/takuseno/minerva",
      author="Takuma Seno",
      author_email="takuma.seno@gmail.com",
      license="MIT License",
      classifiers=["Development Status :: 4 - Beta",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Education",
                   "Intended Audience :: Science/Research",
                   "Topic :: Scientific/Engineering",
                   "Topic :: Scientific/Engineering :: Artificial Intelligence",
                   "Programming Language :: Python :: 3.6",
                   "Programming Language :: Python :: 3.7",
                   "Programming Language :: Python :: 3.8",
                   "Programming Language :: Python :: Implementation :: CPython",
                   "Operating System :: POSIX :: Linux",
                   "Operating System :: MacOS :: MacOS X"],
      install_requires=["d3rlpy==0.70",
                        "Flask",
                        "flask_sqlalchemy",
                        "flask-marshmallow",
                        "Flask-Migrate",
                        "SQLAlchemy",
                        "marshmallow-sqlalchemy",
                        "click",
                        "Pillow"],
      packages=["minerva",
                "minerva.controllers",
                "minerva.models"],
      package_data={'minerva': ['dist/*']},
      python_requires=">=3.5.0",
      include_package_data=True,
      entry_points={'console_scripts': ['minerva=minerva.index:cli']})



import os
import uuid
import json
import base64
from io import BytesIO

import werkzeug
from flask import Blueprint, request, jsonify

from .generator import generate_for_model
from ..config import get_config
from ..dataset import import_csv_as_mdp_dataset, convert_ndarray_to_image
from ..models.dataset import Dataset, DatasetSchema

dataset_route = Blueprint('dataset', __name__)

generate_for_model(dataset_route, Dataset, DatasetSchema)


@dataset_route.route('/upload', methods=['POST'])
def upload_dataset():
    # validation
    if 'dataset' not in request.files:
        return jsonify({'status': 'dataset is empty'}), 400

    # save file
    file = request.files['dataset']
    file_name = werkzeug.utils.secure_filename(file.filename)
    file_path = os.path.join(get_config('UPLOAD_DIR'), file_name)
    file.save(file_path)

    # save as MDPDataset
    is_image = request.form.get('is_image') == 'true'

    # save image files
    if is_image:
        total_images = int(request.form.get('total_images'))
        for i in range(total_images):
            image_file = request.files['image_%d' % i]
            image_name = os.path.basename(image_file.filename)
            image_name = werkzeug.utils.secure_filename(image_name)
            image_path = os.path.join(get_config('UPLOAD_DIR'), image_name)
            image_file.save(image_path)

    mdp_dataset = import_csv_as_mdp_dataset(file_path, image=is_image)
    dataset_name = str(uuid.uuid1()) + '.h5'
    dataset_path = os.path.join(get_config('DATASET_DIR'), dataset_name)
    mdp_dataset.dump(dataset_path)

    # get dataset size
    data_size = os.path.getsize(dataset_path)
    episode_size = len(mdp_dataset)
    step_size = sum(map(len, mdp_dataset))

    # compute statistics
    stats = mdp_dataset.compute_stats()
    stats['observation_shape'] = mdp_dataset.get_observation_shape()
    stats['action_size'] = mdp_dataset.get_action_size()
    # handle ndarray serialization
    stats_json = json.dumps(jsonify(stats).json)

    # insert record
    dataset = Dataset.create(file_name, dataset_name, episode_size, step_size,
                             data_size, is_image,
                             mdp_dataset.is_action_discrete(), stats_json)

    # return json
    return jsonify(DatasetSchema().dump(dataset))


@dataset_route.route('/<dataset_id>/example', methods=['GET'])
def get_example_vector_observation(dataset_id):
    dataset = Dataset.get(dataset_id, raise_404=True)

    # take care of computational cost
    mdp_dataset = dataset.load_mdp_dataset()

    if dataset.is_image:
        # take first 3 samples
        ndarrays = mdp_dataset.observations[:3]

        observations = []
        for ndarray in ndarrays:
            image = convert_ndarray_to_image(ndarray)

            # encode image to base64
            buffer = BytesIO()
            image.save(buffer, format='PNG')
            encoded_image = base64.b64encode(buffer.getvalue())

            # return in string
            observations.append(encoded_image.decode().replace("'", ''))
    else:
        # take first 100 samples
        n_steps = min(100, mdp_dataset.observations.shape[0])
        observations = mdp_dataset.observations[:n_steps]

    return jsonify({'observations': observations})



# pylint: disable=no-name-in-module

from d3rlpy.dataset import MDPDataset
from d3rlpy.algos import create_algo
from d3rlpy.augmentation import DrQPipeline, create_augmentation
from d3rlpy.metrics.scorer import td_error_scorer
from d3rlpy.metrics.scorer import discounted_sum_of_advantage_scorer
from d3rlpy.metrics.scorer import average_value_estimation_scorer
from d3rlpy.metrics.scorer import value_estimation_std_scorer
from d3rlpy.metrics.scorer import continuous_action_diff_scorer
from d3rlpy.metrics.scorer import discrete_action_match_scorer
from d3rlpy.metrics.scorer import initial_state_value_estimation_scorer
from d3rlpy.metrics.scorer import soft_opc_scorer
from sklearn.model_selection import train_test_split


def _get_scorers(discrete_action, dataset_stats):
    scorers = {}
    scorers['td_error'] = td_error_scorer
    scorers['discounted_sum_of_advantage'] = discounted_sum_of_advantage_scorer
    scorers['value_scale'] = average_value_estimation_scorer
    scorers['value_standard_deviation'] = value_estimation_std_scorer
    scorers['initial_state_value'] = initial_state_value_estimation_scorer
    scorers['soft_opc'] = soft_opc_scorer(0.8 * dataset_stats['return']['max'])
    if discrete_action:
        scorers['action_match'] = discrete_action_match_scorer
    else:
        scorers['action_difference'] = continuous_action_diff_scorer
    return scorers


def train(algo_name,
          params,
          dataset_path,
          experiment_name=None,
          logdir='d3rlpy_logs'):
    # prepare dataset
    dataset = MDPDataset.load(dataset_path)
    train_data, test_data = train_test_split(dataset, test_size=0.2)

    # get dataset statistics
    stats = dataset.compute_stats()

    # evaluate
    scorers = _get_scorers(dataset.is_action_discrete(), stats)

    # data augmentation setting
    if 'augmentation' in params:
        aug_names = params['augmentation']
        n_mean = params['n_augmentations']
        augmentations = [create_augmentation(name) for name in aug_names]
        augmentation = DrQPipeline(augmentations, n_mean=n_mean)
        params['augmentation'] = augmentation

    # add action scaler if continuous action-space
    if not dataset.is_action_discrete():
        params['action_scaler'] = 'min_max'

    # train
    algo = create_algo(algo_name, dataset.is_action_discrete(), **params)
    algo.fit(train_data,
             n_epochs=params['n_epochs'],
             eval_episodes=test_data,
             scorers=scorers,
             experiment_name=experiment_name,
             with_timestamp=False,
             logdir=logdir,
             save_interval=1,
             show_progress=False,
             tensorboard=False)

    return True
